﻿namespace SharpScript
{
	#region Using Directives

	using System;
	using System.Collections.Generic;
	using System.Linq;
	using System.Text;
	using Menees;

	#endregion

	internal class ScriptDirectives
	{
		#region Private Data Members

		private List<string> references = new List<string>();
		private List<string> outputFiles = new List<string>();
		private List<string> options = new List<string>();

		#endregion

		#region Constructors

		public ScriptDirectives(string[] scriptLines, ScriptTypeProvider stp, ScriptParameters scriptParams)
		{
			int numLines = scriptLines.Length;
			for (int lineIndex = 0; lineIndex < numLines; lineIndex++)
			{
				string line = scriptLines[lineIndex].TrimStart();
				StringComparison comparison = stp.CaseSensitive ? StringComparison.CurrentCulture : StringComparison.CurrentCultureIgnoreCase;

				string directiveName, directiveValue;
				if (GetDirectiveNameValue(line, stp, comparison, out directiveName, out directiveValue))
				{
					switch (stp.CaseSensitive ? directiveName : directiveName.ToLower())
					{
						case "reference":
							string asmName = GetReferenceName(directiveValue, Properties.Resources.InvalidScriptReference, lineIndex);
							this.references.Add(asmName);
							break;

						case "com_reference":
							string tlbName = GetReferenceName(directiveValue, Properties.Resources.InvalidScriptComReference, lineIndex);
							this.AddComReference(tlbName);
							break;

						case "option":
							if (string.IsNullOrEmpty(directiveValue))
							{
								ThrowDirectiveException(Properties.Resources.InvalidScriptOptionDirective, lineIndex);
							}

							this.options.Add(directiveValue);
							break;

						case "debug":
							if (!string.IsNullOrEmpty(directiveValue))
							{
								ThrowDirectiveException(Properties.Resources.InvalidScriptDebugDirective, lineIndex);
							}

							scriptParams.Debug = true;
							break;
					}
				}
			}
		}

		#endregion

		#region Public Properties

		public IEnumerable<string> References
		{
			get { return this.references; }
		}

		public IEnumerable<string> OutputFiles
		{
			get { return this.outputFiles; }
		}

		public IEnumerable<string> CompilerOptions
		{
			get { return this.options; }
		}

		#endregion

		#region Private Methods

		private static bool GetDirectiveNameValue(
			string line,
			ScriptTypeProvider stp,
			StringComparison comparison,
			out string directiveName,
			out string directiveValue)
		{
			bool result = false;
			directiveName = string.Empty;
			directiveValue = string.Empty;

			if (line.StartsWith(stp.ReferenceLineCommentDelimiter, comparison))
			{
				line = line.Remove(0, stp.ReferenceLineCommentDelimiter.Length).TrimStart();
				if (line.StartsWith("#", comparison) && line.Length > 1)
				{
					int spacePos = line.IndexOfAny(new char[] { ' ', '\t' }, 1);
					if (spacePos >= 0)
					{
						directiveName = line.Substring(1, spacePos - 1);
						if (spacePos + 1 < line.Length)
						{
							directiveValue = line.Substring(spacePos + 1);
						}
					}
					else
					{
						directiveName = line.Substring(1);
					}

					result = true;
				}
			}

			return result;
		}

		private static string GetReferenceName(string directiveValue, string errorMsgFmt, int line)
		{
			string referenceName = null;
			directiveValue = directiveValue.Trim();
			if (directiveValue.StartsWith("\"") && directiveValue.EndsWith("\""))
			{
				referenceName = directiveValue.Substring(1, directiveValue.Length - 2);
			}

			if (!string.IsNullOrEmpty(referenceName))
			{
				return referenceName;
			}
			else
			{
				ThrowDirectiveException(errorMsgFmt, line);
				return null;
			}
		}

		private static void ThrowDirectiveException(string errorMsgFmt, int line)
		{
			string message = string.Format(errorMsgFmt, line + 1);
			throw Exceptions.Log(new ScriptException(message));
		}

		private void AddComReference(string tlbName)
		{
			string namespaceName = null;

			// See if the reference contains a namespace.
			string[] names = tlbName.Split(',', ';');
			if (names.Length == 2)
			{
				tlbName = names[0];
				namespaceName = names[1].Trim();
			}

			TypeLibImporter importer = new TypeLibImporter(tlbName, namespaceName);

			// Add all of the interop assemblies as references.
			string[] interopAssemblies = importer.GetInteropAssemblies();
			this.references.AddRange(interopAssemblies);

			// Add any generated files to the output files.
			string[] outputFiles = importer.GetOutputFiles();
			this.outputFiles.AddRange(outputFiles);
		}

		#endregion
	}
}
