namespace SharpScript
{
	#region Using Directives

	using System;
	using System.CodeDom.Compiler;
	using System.Collections.Generic;
	using System.Collections.Specialized;
	using System.Diagnostics;
	using System.Diagnostics.CodeAnalysis;
	using System.IO;
	using System.Linq;
	using System.Reflection;
	using System.Runtime.InteropServices;
	using System.Text;
	using System.Threading;
	using System.Threading.Tasks;
	using Menees;
	using Microsoft.Win32;

	#endregion

	internal sealed class ScriptHandler : CrossAppDomainObject
	{
		#region Private Data Members

		private static readonly string[] StandardReferences =
			{
				"System.dll",
				"System.Data.dll",
				"System.Xml.dll",
				"System.Core.dll",
				"System.Data.DataSetExtensions.dll",
				"System.Xml.Linq.dll",
			};

		private static readonly string[] GuiReferences =
			{
				"System.Windows.Forms.dll",
				"System.Drawing.dll",
				"System.Deployment.dll",
				"PresentationCore.dll",
				"PresentationFramework.dll",
				"WindowsBase.dll",
				"System.Xaml.dll",
			};

		private ScriptParameters scriptParameters;
		private CompilerResults results;

		#endregion

		#region Constructor

		public ScriptHandler()
		{
			// Before we do anything else, we need to get setup to handle assembly Load
			// failures.  This is necessary since this object is the first thing created
			// in this AppDomain, and we created it using AppDomain.CreateInstanceFrom.
			// That causes this assembly to be loaded into the "LoadFrom" context.  As
			// soon as the Execute method is called, the .NET loader will need to get
			// this assembly into the "Load" context, so it can marshal the ScriptParameters
			// class correctly.  The only way to do that and have the same Assembly reference
			// used in both contexts is to use the AssemblyResolve event.
			AppDomain.CurrentDomain.AssemblyResolve += new ResolveEventHandler(this.CurrentDomain_AssemblyResolve);
		}

		#endregion

		#region Public Methods

		public static void CleanUpOutputFiles(string[] outputFiles)
		{
			if (outputFiles != null)
			{
				Parallel.ForEach(
					outputFiles,
					fileName =>
					{
						if (File.Exists(fileName))
						{
							// This will NOT throw an exception if the file can't be deleted.
							FileUtility.TryDeleteFile(fileName);
						}
					});
			}
		}

		[SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes",
			Justification = "We don't want an exception to escape the Compile step because it might not cross the AppDomain boundary.")]
		public string[] Compile(ScriptParameters parameters, out string exceptionMessage)
		{
			exceptionMessage = string.Empty;
			try
			{
				// Kick this task off first, so it can run while we prep the compiler and read the script.
				Task<IEnumerable<string>> assemblyFoldersTask = Task.Factory.StartNew(() => CacheAssemblyFolders());

				this.scriptParameters = parameters;
				string fileName = this.scriptParameters.FileName;

				// Determine the script type based on the extension.
				ScriptTypeProvider stp = ScriptTypeProvider.GetProviderType(fileName, parameters);

				// Create the correct compiler in parallel while we read the script.
				Task<CodeDomProvider> compilerTask = Task.Factory.StartNew(() => stp.CreateCompiler());

				// Split the script into lines.  We need this to look for
				// script directives and possibly to show the line for
				// each error.
				string[] scriptLines = File.ReadAllLines(fileName);

				// Read all the script directives like #reference, #option, etc.
				ScriptDirectives directives = new ScriptDirectives(scriptLines, stp, this.scriptParameters);

				// Configure the compiler and add all assembly references.
				CompilerParameters compileParams = this.SetupCompilerOptions(stp, directives);
				IEnumerable<string> assemblyFolders = assemblyFoldersTask.Result;
				this.AddAssemblyReferences(compileParams, directives, assemblyFolders, stp.ScriptType == ScriptType.CSharp);

				// Compile the script.
				CodeDomProvider compiler = compilerTask.Result;
				this.results = compiler.CompileAssemblyFromFile(compileParams, fileName);

				// Get all the output files names.
				string[] outputFiles = this.GetOutputFileNames(directives);

				// Throw an exception if there were errors or warnings.
				this.HandleCompileErrors(scriptLines, outputFiles);

				// Return all the output file names, so the caller can clean
				// them up after our AppDomain unloads.
				return outputFiles;
			}
			catch (Exception ex)
			{
				exceptionMessage = ex.Message;
				return null;
			}
		}

		[SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes",
			Justification = "We don't want an exception to escape the Execute step because it might not cross the AppDomain boundary.")]
		public int Execute(out string exceptionMessage)
		{
			exceptionMessage = string.Empty;
			try
			{
				if (this.results == null)
				{
					Debug.Assert(false, "Compile should have been called before Execute.");

					// This doesn't use a string resource because it should never get hit.
					throw Exceptions.Log(new ScriptException("INTERNAL ERROR: Execute called before successful compile."));
				}

				// Initialize the static "Script" class since we're about to execute.
				Script.Initialize(this.scriptParameters);

				// Run the script.
				int result = this.ExecuteAssembly();

				// Try to return the best ExitCode for the process.
				if (result != 0)
				{
					return result;
				}
				else
				{
					return Environment.ExitCode;
				}
			}
			catch (Exception ex)
			{
				exceptionMessage = ex.Message;
				return (int)ExitCode.Exception;
			}
		}

		#endregion

		#region Private Methods

		private static IEnumerable<string> CacheAssemblyFolders()
		{
			List<string> result = new List<string>();

			// The main .NET reference assemblies are always in the 32-bit Program Files directory.
			// (e.g., %ProgramFiles%\Reference Assemblies\Microsoft\Framework\.NETFramework\v4.0).
			// http://msdn.microsoft.com/en-us/library/ee395432.aspx
			// http://blogs.msdn.com/jgoldb/archive/2010/04/12/what-s-new-in-net-framework-4-client-profile-rtm.aspx
			Version dotNetVersion = Environment.Version;
			string programFiles = Environment.ExpandEnvironmentVariables(Environment.Is64BitProcess ?
				"%ProgramFiles(x86)%" : "%ProgramFiles%");
			string baseReferencePath = Path.Combine(programFiles, @"Reference Assemblies\Microsoft\Framework\.NETFramework\");

			// .NET 4.5-4.6 still report as 4.0.30319.x from Environment.Version, but since we're interested in the latest 4.x,
			// we need to look for their reference assemblies first so scripts can refer to any 4.x types and members.
			if (dotNetVersion.Major == 4 && dotNetVersion.Minor == 0)
			{
				// Look through the v4.x.y versions in order from newest to oldest.
				foreach (string version in new[] { "v4.6", "v4.5.2", "v4.5.1", "v4.5" })
				{
					string referencePath = baseReferencePath + version;
					if (Directory.Exists(referencePath))
					{
						result.Add(referencePath);
						break;
					}
				}

				// Note: There is no "Client Profile" version of .NET 4.5.
			}

			if (result.Count == 0)
			{
				string referencePath = baseReferencePath + string.Format("v{0}.{1}", dotNetVersion.Major, dotNetVersion.Minor);
				if (Directory.Exists(referencePath))
				{
					result.Add(referencePath);

					// If only the client profile is installed, then the base reference path may not
					// contain any assemblies. So add the client profile directory too if it exists.
					referencePath += @"\Profile\Client";
					if (Directory.Exists(referencePath))
					{
						result.Add(referencePath);
					}
				}
				else
				{
					// We're probably on a system that doesn't have the .NET SDK or VS installed, so the
					// reference assemblies aren't available.  In that case, we'll just point to the runtime
					// directories we need to satisfy the standard assembly references we add.
					string runtimeDirectory = RuntimeEnvironment.GetRuntimeDirectory();
					result.Add(runtimeDirectory);
					string wpfDirectory = Path.Combine(runtimeDirectory, "WPF");
					if (Directory.Exists(wpfDirectory))
					{
						result.Add(wpfDirectory);
					}
				}
			}

			// Look for other assembly folders configured for the current version of .NET.
			string runtimeAsmFoldersEx = string.Format(
				@"Software\Microsoft\.NETFramework\v{0}.{1}.{2}\AssemblyFoldersEx",
				dotNetVersion.Major,
				dotNetVersion.Minor,
				dotNetVersion.Build);

			// Cache the folders in parallel, but add them to the result list with user entries first.
			Task<IList<string>> userTask = Task.Factory.StartNew(() => CacheAssemblyFolders(runtimeAsmFoldersEx, Registry.CurrentUser));
			Task<IList<string>> machineTask = Task.Factory.StartNew(() => CacheAssemblyFolders(runtimeAsmFoldersEx, Registry.LocalMachine));
			result.AddRange(userTask.Result);
			result.AddRange(machineTask.Result);

			return result;
		}

		private static IList<string> CacheAssemblyFolders(string runtimeAsmFoldersEx, RegistryKey baseKey)
		{
			IList<string> result = new List<string>();

			using (RegistryKey foldersKey = baseKey.OpenSubKey(runtimeAsmFoldersEx))
			{
				if (foldersKey != null)
				{
					string[] folders = foldersKey.GetSubKeyNames();
					foreach (string folderKeyName in folders)
					{
						using (RegistryKey folderKey = foldersKey.OpenSubKey(folderKeyName))
						{
							if (folderKey != null)
							{
								string folder = (string)folderKey.GetValue(string.Empty);
								if (!string.IsNullOrEmpty(folder) && Directory.Exists(folder))
								{
									result.Add(folder);
								}
							}
						}
					}
				}
			}

			return result;
		}

		private static ApartmentState GetRequiredApartmentState(MethodInfo main)
		{
			// See if we need to run this on an STA thread.  Some scripts need
			// this if they're going to invoke STA objects like OpenFileDialog.
			// If they need an STA thread, then they must mark the Main method
			// with the [STAThread] attribute.
			ApartmentState aptState = ApartmentState.MTA;
			try
			{
				if (main.GetCustomAttributes(typeof(STAThreadAttribute), false).Length > 0)
				{
					aptState = ApartmentState.STA;
				}
			}
			catch (CustomAttributeFormatException)
			{
				// JScript 8.0 generates invalid custom attributes, so Reflection's
				// GetCustomAttributes method throws this exception.  Thus there's
				// no way for a JScript SharpScript to run in an STA thread.  However,
				// the JScript compiler generates its own wrapper around the global
				// code, so there's no syntax for specifying the [STAThread] attribute
				// on the real JScript entry point anyway.
				//
				// I logged this with Microsoft on 12/2/2007, but they closed it as "Won't Fix".
				// http://connect.microsoft.com/VisualStudio/feedback/details/314229/jscript-assemblies-dont-support-getcustomattributes
			}

			return aptState;
		}

		private static string GetSharpScriptCommonPath()
		{
			string result = Assembly.GetExecutingAssembly().Location;
			return result;
		}

		private static void AddAssemblyReferences(StringCollection references, IEnumerable<string> assemblyFolders, IEnumerable<string> fileNames)
		{
			Parallel.ForEach(
				fileNames,
				fileName =>
				{
					string qualifiedName = QualifyFromAssemblyFolders(fileName, assemblyFolders);
					lock (references)
					{
						references.Add(qualifiedName);
					}
				});
		}

		private static string QualifyFromAssemblyFolders(string asmRef, IEnumerable<string> assemblyFolders)
		{
			// This isn't parallelized because we want to match in order
			// (i.e., earlier folders are preferable to later folders).
			foreach (string folder in assemblyFolders)
			{
				string qualifiedRef = Path.Combine(folder, asmRef);
				if (File.Exists(qualifiedRef))
				{
					return qualifiedRef;
				}
			}

			return asmRef;
		}

		private CompilerParameters SetupCompilerOptions(ScriptTypeProvider stp, ScriptDirectives directives)
		{
			CompilerParameters compileParams = new CompilerParameters();

			// We have to do this so we'll have an EntryPoint.
			compileParams.GenerateExecutable = true;

			// If we're including debug information, then we need to generate
			// the file on disk and not just in memory.
			bool debug = this.scriptParameters.Debug;
			compileParams.GenerateInMemory = !debug;
			compileParams.IncludeDebugInformation = debug;
			if (debug)
			{
				// I'm appending ".exe" rather than changing the file's extension.
				// This makes it possible to have two script files with different
				// extensions running in the same directory without clobbering each
				// other's output.  That can happen in the Samples directory.
				compileParams.OutputAssembly = this.scriptParameters.FileName + ".exe";
			}

			// Make the script options as strict as possible.  If people want
			// weak checking, they should use WScript.
			compileParams.TreatWarningsAsErrors = true;
			compileParams.WarningLevel = 4;

			// Set the custom compile options for the current build mode.
			StringBuilder options = new StringBuilder(compileParams.CompilerOptions);

			// Set the executable to the correct target type.
			// This isn't strictly necessary, but if the temporary
			// exe gets left out there and a user tries to run it,
			// I'd like it to have the correct type.
			if (stp.SupportsTargetSwitch)
			{
				if (this.scriptParameters.InConsole)
				{
					options.Append(" /target:exe");
				}
				else
				{
					options.Append(" /target:winexe");
				}
			}

			// Get custom compile options for the provider type.
			options.Append(' ');
			if (debug)
			{
				options.Append(stp.DebugCompileOptions);
			}
			else
			{
				options.Append(stp.ReleaseCompileOptions);
			}

			// Add options from any #option directives.
			foreach (string option in directives.CompilerOptions)
			{
				options.Append(' ');
				options.Append(option);
			}

			compileParams.CompilerOptions = options.ToString();

			return compileParams;
		}

		private void AddAssemblyReferences(
			CompilerParameters compileParams,
			ScriptDirectives directives,
			IEnumerable<string> assemblyFolders,
			bool isCSharp)
		{
			// Add the same default assembly references that VS uses for new projects.
			// Most compilers should automatically add mscorlib.dll.
			List<string> referenceNames = new List<string>(StandardReferences.Length + GuiReferences.Length + 1);
			referenceNames.AddRange(StandardReferences);
			if (isCSharp)
			{
				referenceNames.Add("Microsoft.CSharp.dll");
			}

			if (!this.scriptParameters.InConsole)
			{
				// Give GUI apps the default references for both WinForms and WPF apps.
				referenceNames.AddRange(GuiReferences);
			}

			StringCollection references = compileParams.ReferencedAssemblies;
			AddAssemblyReferences(references, assemblyFolders, referenceNames);

			// Add the SharpScript.Common assembly.
			string sharpScriptCommonPath = GetSharpScriptCommonPath();
			references.Add(sharpScriptCommonPath);

			// Add assembly references from within the file (e.g., //#reference "MyAssembly.dll").
			string scriptBaseDirectory = Path.GetDirectoryName(this.scriptParameters.FileName);
			Parallel.ForEach(
				directives.References,
				asmRef =>
				{
					string qualifiedAsmRef = this.GetQualifiedAssemblyReference(asmRef, scriptBaseDirectory, assemblyFolders);
					lock (references)
					{
						references.Add(qualifiedAsmRef);
					}
				});
		}

		private string GetQualifiedAssemblyReference(string asmRef, string scriptBaseDirectory, IEnumerable<string> assemblyFolders)
		{
			// First expand any environment variable references so assembly refs can be system-independent.
			asmRef = Environment.ExpandEnvironmentVariables(asmRef);

			// If the assembly reference is already fully qualified, then we don't
			// have to do anything else.
			if (Path.IsPathRooted(asmRef))
			{
				return asmRef;
			}

			// See if the file exists under the script base directory.  If so, then
			// we want to use it.  This way we can compile against local "xcopied"
			// assemblies that aren't registered anywhere.
			//
			// Adjust the process's CurrentDirectory and use GetFullPath, so we
			// can correctly resolve relative paths to references.
			string originalCurrentDirectory = Environment.CurrentDirectory;
			Environment.CurrentDirectory = scriptBaseDirectory;
			try
			{
				string qualifiedAsmRef = Path.GetFullPath(asmRef);
				if (File.Exists(qualifiedAsmRef))
				{
					return qualifiedAsmRef;
				}

				// See if it's a shared assembly under the SharpScript folder.
				Environment.CurrentDirectory = this.scriptParameters.SharpScriptDirectory;
				qualifiedAsmRef = Path.GetFullPath(asmRef);
				if (File.Exists(qualifiedAsmRef))
				{
					return qualifiedAsmRef;
				}
			}
			finally
			{
				Environment.CurrentDirectory = originalCurrentDirectory;
			}

			// The assembly name wasn't fully qualified, and it doesn't exist
			// when we qualify it from the script base directory or SharpScript
			// directory.  So we'll see if the assembly exists under any of the
			// configured AssemblyFolders in the registry.
			asmRef = QualifyFromAssemblyFolders(asmRef, assemblyFolders);
			return asmRef;
		}

		private void HandleCompileErrors(string[] scriptLines, string[] outputFiles)
		{
			int numErrors = this.results.Errors.Count;
			if (numErrors > 0)
			{
				StringBuilder sb = new StringBuilder(1024);
				sb.Append(Properties.Resources.CompileErrors);
				sb.Append(":\r\n");

				foreach (CompilerError error in this.results.Errors)
				{
					int lineNumber = error.Line;

					if (lineNumber > 0)
					{
						sb.AppendFormat(Properties.Resources.LineColumn, lineNumber, error.Column);
					}

					// Show the standard error information.
					string type = error.IsWarning ? Properties.Resources.Warning : Properties.Resources.Error;
					sb.AppendFormat("{0} {1}: {2}\r\n", type, error.ErrorNumber, error.ErrorText);

					// Show the line text with each error message.  But check the bounds because
					// I've seen the VB compiler return a line number equal to scriptLines.Length + 3.
					if (lineNumber > 0 && lineNumber <= scriptLines.Length)
					{
						sb.Append('\t');
						string line = scriptLines[lineNumber - 1].Trim();
						sb.Append(line);
						sb.Append("\r\n");
					}

					sb.Append("\r\n");
				}

				// Clean up the output files before we throw since the
				// caller never had a chance to get them.
				CleanUpOutputFiles(outputFiles);

				string errors = sb.ToString();
				throw Exceptions.Log(new ScriptException(errors));
			}
		}

		private string[] GetOutputFileNames(ScriptDirectives directives)
		{
			List<string> lstOutputFiles = new List<string>();

			// Add the assembly.
			string asmPath = this.results.PathToAssembly;
			if (asmPath != null && asmPath.Length > 0)
			{
				lstOutputFiles.Add(asmPath);

				// Add the debug symbols if necessary.  They're never
				// listed as one of the temporary files.
				if (this.scriptParameters.Debug)
				{
					lstOutputFiles.Add(Path.ChangeExtension(asmPath, ".pdb"));
				}
			}

			// Add any temporary files.
			foreach (string fileName in this.results.TempFiles)
			{
				if (fileName != null && fileName.Length > 0)
				{
					lstOutputFiles.Add(fileName);
				}
			}

			// Add any output files required by script directives like #com_reference.
			lstOutputFiles.AddRange(directives.OutputFiles);

			return lstOutputFiles.ToArray();
		}

		[SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes",
			Justification = "We don't want an exception to escape the thread's entry point method.")]
		private int ExecuteAssembly()
		{
			Assembly compiledAssembly;
			try
			{
				// Get the main entry point to the new assembly.
				compiledAssembly = this.results.CompiledAssembly;
			}
			catch (BadImageFormatException ex)
			{
				string extra = "If the script is using the /platform compiler option, then it can " +
					"only target AnyCpu or " + (Environment.Is64BitProcess ? "x64." : "x86");
				throw new BadImageFormatException(ex.Message + "\r\n" + extra, ex);
			}

			MethodInfo main = compiledAssembly.EntryPoint;
			if (main == null)
			{
				throw new InvalidProgramException("The script's entry point (e.g., a Main method) could not be found.");
			}

			// Main's parameter can be void or string[].  We need to handle both cases.
			object[] mainArgs = null;
			if (main.GetParameters().Length == 1)
			{
				mainArgs = new object[] { this.scriptParameters.Arguments };
			}

			// These will be set the by the thread's anonymous delegate.
			Exception exception = null;
			int result = 0;

			// Always execute in a new worker thread so we can control its apartment model.
			// Remember: This code won't execute until Thread.Start is called!
			Thread thread = new Thread(delegate()
				{
					// Run the Main method.
					try
					{
						object invokeResult = main.Invoke(null, mainArgs);

						// If it returns a value, then pass it back.  Otherwise, use the Environment's ExitCode.
						if (main.ReturnType == typeof(int))
						{
							result = (int)invokeResult;
						}
						else
						{
							result = Environment.ExitCode;
						}
					}
					catch (TargetInvocationException ex)
					{
						// In .NET 2.0, Main.Invoke now throws a TargetInvocationException
						// if the Main method throws an exception.  But the caller doesn't need
						// to see "Exception has been thrown by the target of an invocation."
						// every time.  The caller needs to see the original exception that got
						// thrown by the script.
						if (ex.InnerException != null)
						{
							exception = ex.InnerException;
						}
						else
						{
							exception = ex;
						}
					}
					catch (Exception ex)
					{
						exception = ex;
					}
				});

			// Determine the apartment state that the thread needs to use.
			ApartmentState aptState = GetRequiredApartmentState(main);
			thread.SetApartmentState(aptState);

			// Run the thread and wait for it to finish.
			thread.Start();
			thread.Join();

			// Check if we need to rethrow an exception from the thread function.
			if (exception != null)
			{
				// In debug runs, we'll rethrow a message containing all the exception detail
				// such as type, stack trace, message, etc.  Release runs just get the message.
				string message = this.scriptParameters.Debug ? exception.ToString() : exception.Message;
				throw Exceptions.Log(new ScriptException(message));
			}

			return result;
		}

		private Assembly CurrentDomain_AssemblyResolve(object sender, ResolveEventArgs args)
		{
			// If we couldn't find the assembly in the "Load" context, look through
			// all assemblies in the current AppDomain because the assembly may
			// be in the "LoadFrom" context.
			Assembly[] assemblies = AppDomain.CurrentDomain.GetAssemblies();
			foreach (Assembly assembly in assemblies)
			{
				if (assembly.FullName == args.Name)
				{
					return assembly;
				}
			}

			// If we still can't find the assembly, see if it exists in the base SharpScript folder.
			// This allows scripts to use the Menees.* assemblies as well as any other shared
			// assemblies a user wants to deploy into the base SharpScript folder.
			string sharpScriptFolder = Path.GetDirectoryName(GetSharpScriptCommonPath());
			int commaIndex = args.Name.IndexOf(',');
			string fileNameNoExt = commaIndex < 0 ? args.Name : args.Name.Substring(0, commaIndex);
			string dllName = Path.Combine(sharpScriptFolder, fileNameNoExt + ".dll");
			if (File.Exists(dllName))
			{
				Assembly assembly = Assembly.LoadFrom(dllName);
				return assembly;
			}

			return null;
		}

		#endregion
	}
}
