namespace SharpScript
{
	#region Using Directives

	using System;
	using System.CodeDom.Compiler;
	using System.Collections.Generic;
	using System.Diagnostics;
	using System.IO;
	using System.Reflection;
	using System.Xml;
	using Menees;

	#endregion

	internal sealed class ScriptTypeProvider
	{
		#region Private Data Members

		private static List<ScriptTypeProvider> allProviders = new List<ScriptTypeProvider>();
		private static bool initialized;

		private Type providerType;
		private string[] extensions;
		private string debugCompileOptions;
		private string releaseCompileOptions;
		private string referenceLineCommentDelimiter;
		private bool supportsTargetSwitch;
		private bool caseSensitive;
		private ScriptType scriptType;

		#endregion

		#region Constructors

		public ScriptTypeProvider(
			Type providerType,
			string[] extensions,
			string debugCompileOptions,
			string releaseCompileOptions,
			string referenceLineCommentDelimiter,
			bool supportsTargetSwitch,
			bool caseSensitive,
			ScriptType scriptType)
		{
			this.providerType = providerType;
			this.debugCompileOptions = debugCompileOptions;
			this.releaseCompileOptions = releaseCompileOptions;
			this.referenceLineCommentDelimiter = referenceLineCommentDelimiter;
			this.supportsTargetSwitch = supportsTargetSwitch;
			this.caseSensitive = caseSensitive;
			this.scriptType = scriptType;

			int numExtensions = extensions.Length;
			this.extensions = new string[numExtensions];
			for (int i = 0; i < numExtensions; i++)
			{
				string extension = extensions[i].ToLower().Trim();
				if (!extension.StartsWith("."))
				{
					extension = "." + extension;
				}

				this.extensions[i] = extension;
			}
		}

		#endregion

		#region Public Properties

		public string[] Extensions
		{
			get { return this.extensions; }
		}

		public string DebugCompileOptions
		{
			get { return this.debugCompileOptions; }
		}

		public string ReleaseCompileOptions
		{
			get { return this.releaseCompileOptions; }
		}

		public string ReferenceLineCommentDelimiter
		{
			get { return this.referenceLineCommentDelimiter; }
		}

		public bool SupportsTargetSwitch
		{
			get { return this.supportsTargetSwitch; }
		}

		public bool CaseSensitive
		{
			get { return this.caseSensitive; }
		}

		public ScriptType ScriptType
		{
			get { return this.scriptType; }
		}

		#endregion

		#region Public Methods

		public static ScriptTypeProvider GetProviderType(string fileName, ScriptParameters parameters)
		{
			if (!initialized)
			{
				Initialize(parameters);
			}

			string fileExtension = Path.GetExtension(fileName).ToLower();

			foreach (ScriptTypeProvider provider in allProviders)
			{
				foreach (string typeExtension in provider.Extensions)
				{
					if (typeExtension == fileExtension)
					{
						return provider;
					}
				}
			}

			// If we get here, then the extension is an unsupported type.
			string message = string.Format(Properties.Resources.UnsupportedExtension, fileExtension);
			throw Exceptions.Log(new ScriptException(message));
		}

		public CodeDomProvider CreateCompiler()
		{
			CodeDomProvider result;
			if (this.providerType != null)
			{
				result = (CodeDomProvider)Activator.CreateInstance(this.providerType);
			}
			else
			{
				// Determine the current .NET version.  Due to the app.config file's
				// <supportedRuntime> element, it may not be the same version of
				// .NET that SharpScript was built against.  It could be newer.
				Version dotNetVersion = Environment.Version;

				// For .NET 3.5 (actually in .NET 2.0 SP1), Microsoft added special new constructors
				// to the C# and VB CodeDomProvider classes, and that's the only programmatic
				// way to force them to use specific version language features.  So I'm hardcoding
				// these constructor calls rather than adding even more configurable options for
				// special constructor invocation.
				Dictionary<string, string> providerOptions = new Dictionary<string, string>();
				providerOptions.Add("CompilerVersion", string.Format("v{0}.{1}", dotNetVersion.Major, dotNetVersion.Minor));

				if (this.scriptType == ScriptType.VB)
				{
					result = new Microsoft.VisualBasic.VBCodeProvider(providerOptions);
				}
				else
				{
					result = new Microsoft.CSharp.CSharpCodeProvider(providerOptions);
				}
			}

			return result;
		}

		#endregion

		#region Private Methods

		private static void Initialize(ScriptParameters parameters)
		{
			// Put in the pre-known provider types first so no one can change them
			// from a config file.  We always search the list in order, so we'll
			// match the standard extensions first.  If a user really doesn't like
			// the behavior of one of these, they can define a custom extension
			// and use that with a new provider type with the options they want.
			allProviders.Add(new ScriptTypeProvider(
				null,
				new string[] { ".scs", ".cs" },
				"/define:TRACE;DEBUG /checked",
				"/define:TRACE /optimize",
				"//",
				true,
				true,
				ScriptType.CSharp));

			string visualBasicImports = "/imports:Microsoft.VisualBasic,System,System.Collections,System.Diagnostics," +
				"System.Data,SharpScript,System.Linq,System.Xml.Linq";
			allProviders.Add(new ScriptTypeProvider(
				null,
				new string[] { ".svb", ".vb" },
				"/define:TRACE=True,DEBUG=True /optionexplicit+ /optionstrict+ /optioninfer+ " + visualBasicImports,
				"/define:TRACE=True /optionexplicit+ /optionstrict+ /optioninfer+ /optimize /removeintchecks+ " + visualBasicImports,
				"'",
				true,
				false,
				ScriptType.VB));

			string fileName = parameters.CustomConfigFile;
			if (!string.IsNullOrEmpty(fileName))
			{
				XmlDocument doc = new XmlDocument();
				doc.Load(fileName);

				XmlNodeList lstProviders = doc.SelectNodes("//ScriptTypeProvider");
				foreach (XmlNode node in lstProviders)
				{
					XmlElement provider = node as XmlElement;
					if (provider != null)
					{
						// Get the required provider data.
						string providerTypeName = GetRequiredAttribute(provider, "ProviderTypeName");
						string providerAssemblyName = GetRequiredAttribute(provider, "ProviderAssemblyName");
						string referenceLineCommentDelimiter = GetRequiredAttribute(provider, "ReferenceLineCommentDelimiter");
						string extensionsText = GetRequiredAttribute(provider, "Extensions");
						string[] extensions = extensionsText.Split(';', ',', ' ');

						// Get the optional provider data.
						string debugCompileOptions = provider.GetAttribute("DebugCompileOptions");
						string releaseCompileOptions = provider.GetAttribute("ReleaseCompileOptions");
						bool supportsTargetSwitch = string.Compare(provider.GetAttribute("SupportsTargetSwitch"), "true") == 0;
						bool caseSensitive = string.Compare(provider.GetAttribute("CaseSensitive"), "true") == 0;

						// Resolve the provider type.
						Assembly providerAssembly = Assembly.Load(providerAssemblyName);
						Type providerType = providerAssembly.GetType(providerTypeName);

						// Add the new provider type.
						ScriptTypeProvider stp = new ScriptTypeProvider(
							providerType,
							extensions,
							debugCompileOptions,
							releaseCompileOptions,
							referenceLineCommentDelimiter,
							supportsTargetSwitch,
							caseSensitive,
							ScriptType.Custom);
						allProviders.Add(stp);
					}
				}
			}

			initialized = true;
		}

		private static string GetRequiredAttribute(XmlElement node, string attribute)
		{
			string value = node.GetAttribute(attribute);
			if (value == null || value.Length == 0)
			{
				string message = string.Format(Properties.Resources.RequiredAttribute, attribute);
				throw new ArgumentException(message);
			}

			return value;
		}

		#endregion
	}
}
